from . import drawing

def render(out, named_tensors, filename, cull = [], draw_fn = drawing.draw_all, fmt = 'pdf', **draw_kwargs):
    G = None
    if 'torch' in str(type(out)):
        from .make_graph import traverse_torch
        G = traverse_torch(out, named_tensors)
    else:
        print('Unsupported type. Supported AG libraries: [pytorch, ]')

    if G is None:
        raise Exception('Failed to produce networkx graph from autograd data. See errors above.')

    if len(cull) > 0:
        from .graph_modifiers import remove_nodes
        G = remove_nodes(G, cull)

#    if color_special:
#        from .graph_modifiers import color_edges_on_any_path
#        sources = [n for n, attr in G.nodes(data=True) if attr['group'] == 'param']
#        targets = [n for n, attr in G.nodes(data=True) if attr['group'] == 'state']
#        G = color_edges_on_any_path(G, sources, targets)

    dot = draw_fn(G, **draw_kwargs)
    dot.format = fmt 
    dot.render(filename, cleanup=True)



import networkx as nx

def remove_nodes(G, group):
    # merge all nodes of a certain group type in graph. group can be a single group or a list of groups.
    if isinstance(group, str):
        group = [group]

    def remove_node_and_keep_edges(node):
        preds = list(G.predecessors(node))
        succs = list(G.successors(node))
        for u in preds:
            for v in succs:
                if u != v:
                    G.add_edge(u, v)

    rm_list = []
    for node, attrs in G.nodes(data=True):
        if attrs['group'] in group:
            remove_node_and_keep_edges(node)
            rm_list.append(node)

    G.remove_nodes_from(rm_list)
    return G


def color_edges_on_any_path(G, sources, targets, color="red"):
    """
    Color all edges that lie on at least one path from any node in `sources`
    to any node in `targets` with edge attribute {'color': color}.
    Works for DiGraph and MultiDiGraph.
    """
    # Normalize inputs
    sources = [s for s in sources if s in G]
    targets = [t for t in targets if t in G]
    if not sources or not targets:
        return G  # nothing to do

    # 1) Forward reachability from sources
    forward = set()
    for s in sources:
        forward.add(s)
        forward |= nx.descendants(G, s)

    # 2) Backward reachability to targets (i.e., descendants in the reversed graph)
    R = G.reverse(copy=False)
    backward = set()
    for t in targets:
        backward.add(t)
        backward |= nx.descendants(R, t)

    # 3) Color every edge whose tail is forward-reachable and head can reach a target
    for u, v, data in G.edges(data=True):
        if u in forward and v in backward:
            data["color"] = color

    return G



# Convert autograd tree to a generic network with metadata for nodes/edges.
# I chose to use networkx for the "generic network" model instead of building it 
# by hand since it supports things like removing nodes, finding paths, etc.
# autograd tree can be specified in common languages like Jax, Pytorch, etc.
# SUPPORT NOW: only pytorch. 

from networkx import DiGraph

def traverse_torch(out, named_tensors):
    # Traverse the pytorch AD tree and create a graph.
    # Nodes have meta data, including:
    # - id: a raw identifier for the node.
    # - label: visible name to use when visualizing.
    # - group: one of 'state', 'param', 'none' or 'func'. none represents leaves that are not specified by user or stored tensors for functions.
    #
    # named_tensors should be a dict with entries like 'state_10': (tensor, 'state') specifying maps from names to tuple of tensor and group (one of 'state' or 'param')
    
    G = DiGraph()
    tensor_name = {id(t): name for name, (t, _) in named_tensors.items()}

    new_node = lambda x, label, group: G.add_node(str(id(x)), label=label, group=group)
    new_edge = lambda x1, x2: G.add_edge(str(id(x1)), str(id(x2)))

    # Add explicit nodes for all named tensors that the user specified.
    for name, (t, group) in named_tensors.items():
        new_node(t, f"{name}\n{tuple(t.shape)}", group)

    seen = set()
    def dfs_call(fn): # Recursively called to traverse tree (Depth-First-Search)
        if fn is None or fn in seen:
            return
        seen.add(fn)

        new_node(fn, type(fn).__name__, 'func')

        # Connect saved tensors (optional, dashed)
        for t in getattr(fn, "saved_tensors", []):
            if id(t) not in tensor_name:
                new_node(t, f"saved{tuple(t.shape)}", 'none')
            new_edge(t, fn)

        # Upstream edges
        for nxt, _ in getattr(fn, "next_functions", []):
            if nxt is None:
                continue
            if hasattr(nxt, "variable"):  # AccumulateGrad for a leaf variable
                v = nxt.variable
                # ensure the leaf shows up (named if we have it)
                if id(v) not in tensor_name:
                    new_node(v, f"leaf{tuple(v.shape)}", 'none')
                new_edge(v, fn)
            else:
                new_edge(nxt, fn)
                dfs_call(nxt)

    dfs_call(out.grad_fn)

    # Also connect each named non-leaf tensor to its producing function so labels are obvious
    for (t, _) in named_tensors.values():
        if t.grad_fn is not None:
            new_edge(t.grad_fn, t)

    return G



from graphviz import Digraph

def draw_all(G, color_edges = True, simplify = ['none', 'func'], **glob_attrs):
    dot = Digraph(node_attr={'fontsize': glob_attrs.get('fontsize', '20')})
    dot.attr(rankdir=glob_attrs.get('rankdir', 'LR'))
    dot.edge_attr.update(pen_width=glob_attrs.get('pen_width', '4'))
    dot.attr(ranksep=glob_attrs.get('ranksep', '0.1'), nodesep=glob_attrs.get('nodesep', '0.2'))

    # nodes
    for n, d in G.nodes(data=True):
        group = d['group']
        shape = {'state': 'ellipse', 'param': 'square', 'none': 'ellipse', 'func': 'rectangle'}[group]
        color = {'state': '#8df2b2ff:transparent', 'param': '#8de9f2ff:transparent', 'none': 'grey:transparent', 'func': 'grey:transparent'}[group]
        attrs = {'label': d['label'], 'shape': shape, 'fillcolor': color, 'style': 'filled'}
        if group in simplify: # Just draw a black dot.
            attrs['label'] = ''
            attrs['width'] = '0.1'
            attrs['height'] = '0.1'
            attrs['color'] = 'black'
        dot.node(str(n), **attrs)

    color_cycle = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf'] # Matplotlib default cycler.
    for eidx, (u, v, attr) in enumerate(G.edges(data = True)):
        if 'color' in attr: 
            dot.edge(str(u), str(v), color = attr['color'])
        else:
            dot.edge(str(u), str(v), color = (color_cycle[eidx % len(color_cycle)] if color_edges else 'black'))

    return dot



from setuptools import setup, find_packages

setup(
    name="agviz",
    version="1.0.1",
    description="AutoGrad Visualizer (agviz): Visualize dependencies between states/parameters in autograd tree.",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    author="James Hazelden",
    url="https://github.com/meeree/agviz",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=[
        'graphviz',
        'networkx'
    ],
    python_requires=">=3.2",
)
